using System;
using System.Collections.Generic;
using System.Text;
using System.Expressions;
using System.Query;

namespace MySpecification
{

     public interface IQuickSpecification<T>
    {
        Expression<Predicate<T>> Predicate {get;set;}
        IQuickSpecification<T> And(IQuickSpecification<T> other);
        bool IsSatisfiedBy(T customer);
        IQuickSpecification<T> Not();
        IQuickSpecification<T> Or(IQuickSpecification<T> other);
    }
    
    
    abstract public class AbstractQuickSpecification<T> : IQuickSpecification<T>
    {
        protected Expression<Predicate<T>> predicate;
        public abstract bool IsSatisfiedBy(T value);

        #region IQuickSpecification<T> Members

        public Expression<Predicate<T>> Predicate
        {
            get
            {
                return predicate;
            }
            set
            {
                predicate = value;
            }
        }

        public IQuickSpecification<T> And(IQuickSpecification<T> other)
        { 
            return new QuickAndSpecification<T>(this, other);
        }

        public IQuickSpecification<T> Or(IQuickSpecification<T> other)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public IQuickSpecification<T> Not()
        {
            throw new Exception("The method or operation is not implemented.");
        }
        #endregion
    }
   
    public class QuickSpecification<T> : AbstractQuickSpecification<T>
    {
       
        public QuickSpecification(Expression<Predicate<T>> predicate)
        {
            this.predicate = predicate;
        }

        public override bool IsSatisfiedBy(T value)
        {
            return predicate.Compile().Invoke(value);
        }
    }

    public class QuickAndSpecification<T> : AbstractQuickSpecification<T>
    {
       private IQuickSpecification<T> rhs;
       private IQuickSpecification<T> lhs;
       public QuickAndSpecification(IQuickSpecification<T> rhs, IQuickSpecification<T> lhs)
       {
           this.rhs = rhs;
           this.lhs = lhs;
       }

       public override bool IsSatisfiedBy(T candidate)
       {
           ParameterExpression parameter = Expression.Parameter(typeof(T), "p");
                                 
           var invokedExpression = Expression.Invoke (lhs.Predicate, rhs.Predicate.Parameters.Cast<Expression>());
           Expression<Predicate<T>> andPredicate = Expression.Lambda<Predicate<T>>(Expression.And (rhs.Predicate.Body, invokedExpression), rhs.Predicate.Parameters);
           return andPredicate.Compile().Invoke(candidate);
       }
    }
}
